
class TCP_Flags():
    def __init__(self, integer = 0):
        self.fin_flag = (integer & 1) > 0
        self.syn_flag = (integer & 2) > 0
        self.rst_flag = (integer & 4) > 0
        self.psh_flag = (integer & 8) > 0
        self.ack_flag = (integer & 16) > 0
        self.urg_flag = (integer & 32) > 0
        self.ece_flag = (integer & 64) > 0
        self.cwr_flag = (integer & 128) > 0

    def set_fin_flag(self, value):
        self.fin_flag = value

    def set_syn_flag(self, value):
        self.syn_flag = value

    def set_rst_flag(self, value):
        self.rst_flag = value

    def set_psh_flag(self, value):
        self.psh_flag = value

    def set_ack_flag(self, value):
        self.ack_flag = value

    def set_urg_flag(self, value):
        self.urg_flag = value

    def set_ece_flag(self, value):
        self.ece_flag = value

    def set_cwr_flag(self, value):
        self.cwr_flag = value

    def get_fin_flag(self):
        return self.fin_flag

    def get_syn_flag(self):
        return self.syn_flag

    def get_rst_flag(self):
        return self.rst_flag

    def get_psh_flag(self):
        return self.psh_flag

    def get_ack_flag(self):
        return self.ack_flag

    def get_urg_flag(self):
        return self.urg_flag

    def get_ece_flag(self):
        return self.ece_flag

    def get_cwr_flag(self):
        return self.cwr_flag

    def get_integer(self):
        integer = 0
        if (self.fin_flag):
            integer = integer | 1
        if (self.syn_flag):
            integer = integer | 2
        if (self.rst_flag):
            integer = integer | 4
        if (self.psh_flag):
            integer = integer | 8
        if (self.ack_flag):
            integer = integer | 16
        if (self.urg_flag):
            integer = integer | 32
        if (self.ece_flag):
            integer = integer | 64
        if (self.cwr_flag):
            integer = integer | 128
        return integer




import socket
import struct

from tcp_send_first_segment_last.lib.checksum import calculate_checksum
from tcp_send_first_segment_last.lib.TCP_Flags import TCP_Flags


def print_byte_string(byte_str):
    word = ""
    for byte in byte_str:
        word += "{:02x}".format(byte)
        if len(word) == 4:
            print(word)
            word = ""

class TCP_Segment():
    def __init__(
            self, src_port, dst_port, seq_num, ack_num, flags, data = b""):
        self.src_port = src_port
        self.dst_port = dst_port
        self.seq_num = seq_num
        self.ack_num = ack_num
        self.flags = flags
        self.data = data

    # Python classes can only have 1 constructor
    @classmethod
    def from_bytes(klass, byte_str):
        src_port = int.from_bytes(byte_str[0:2])
        dst_port = int.from_bytes(byte_str[2:4])
        seq_num = int.from_bytes(byte_str[4:8])
        ack_num = int.from_bytes(byte_str[8:12])
        data_offset = (byte_str[12] & 0xF0) >> 4
        flags = TCP_Flags(byte_str[13])
        data = byte_str[(data_offset * 4):]
        # print_byte_string(data)
        return klass(
                src_port, dst_port, seq_num, ack_num, flags, data)

    def get_dst_port(self):
        return self.dst_port

    def get_seq_num(self):
        return self.seq_num

    def get_ack_num(self):
        return self.ack_num

    def get_flags(self):
        return self.flags

    def get_bytes(self, src_addr, dst_addr):
        # Always the same for now
        data_offset_plus_reserved_bits = 5 << 4
        window_size = 65535

        pseudo_header = socket.inet_aton(src_addr)
        pseudo_header += socket.inet_aton(dst_addr)
        pseudo_header += struct.pack(">H", 0x6)
        pseudo_header += struct.pack(">H", 20 + len(self.data))

        tcp_header_top = struct.pack(">H", self.src_port)
        tcp_header_top += struct.pack(">H", self.dst_port)
        tcp_header_top += struct.pack(">I", self.seq_num)
        tcp_header_top += struct.pack(">I", self.ack_num)
        tcp_header_top += struct.pack(">B", data_offset_plus_reserved_bits)
        tcp_header_top += struct.pack(">B", self.flags.get_integer())
        tcp_header_top += struct.pack(">H", window_size)
        # Checksum will go here
        tcp_header_urgent = struct.pack(">H", 0)

        checksum_data = self.data
        if not (len(self.data) % 2) == 0:
            checksum_data += b"\x00"

        checksum = calculate_checksum(
                pseudo_header + tcp_header_top + struct.pack(">H", 0) +
                tcp_header_urgent + checksum_data)

        return tcp_header_top + struct.pack(">H", checksum) + tcp_header_urgent + self.data




import socket
import time

from tcp_send_first_segment_last.lib.disable_auto_rst import disable, cleanup
from tcp_send_first_segment_last.lib.IP_Datagram import IP_Datagram
from tcp_send_first_segment_last.lib.TCP_Segment import TCP_Segment
from tcp_send_first_segment_last.lib.TCP_Flags import TCP_Flags



def get_response(sock, dst_port):
    while True:
        data = sock.recv(1024)
        ip_datagram = IP_Datagram.from_bytes(data)
        tcp_segment = ip_datagram.get_tcp_segment()
        if tcp_segment.get_dst_port() == dst_port:
            return ip_datagram

def reset_connection(
        sock, src_addr, dst_addr, src_port, dst_port):
    # Send Rst packet
    flags = TCP_Flags()
    flags.set_rst_flag(True)
    req_segment = TCP_Segment(src_port, dst_port, 0, 0, flags)
    sock.sendall(req_segment.get_bytes(src_addr, dst_addr))

def establish_connection(
        sock, src_addr, dst_addr, src_port, dst_port):
    # Send Syn packet
    flags = TCP_Flags()
    flags.set_syn_flag(True)
    req_segment = TCP_Segment(src_port, dst_port, 0, 0, flags)
    sock.sendall(req_segment.get_bytes(src_addr, dst_addr))

    # Receive Syn-Ack
    res_dgram = get_response(sock, src_port)
    res_segment = res_dgram.get_tcp_segment()
    flags = res_segment.get_flags()
    if not flags.get_syn_flag():
        seq_num = res_segment.get_ack_num()
        ack_num = res_segment.get_seq_num()
        terminate_connection(
                sock, src_addr, dst_addr, src_port, dst_port, seq_num, ack_num)
        return establish_connection(
                sock, src_addr, dst_addr, src_port, dst_port)

    # Send Ack packet
    flags = TCP_Flags()
    flags.set_ack_flag(True)
    seq_num = res_segment.get_ack_num()
    ack_num = res_segment.get_seq_num() + 1
    req_segment = TCP_Segment(src_port, dst_port, seq_num, ack_num, flags)
    sock.sendall(req_segment.get_bytes(src_addr, dst_addr))

    return (seq_num, ack_num)

def terminate_connection(
        sock, src_addr, dst_addr, src_port, dst_port,
        seq_num, ack_num, fin_ack_received = False):
    if fin_ack_received:
        ack_num += 1

    # Send Fin-Ack packet
    flags = TCP_Flags()
    flags.set_fin_flag(True)
    # For some reason, closing the connection doesn't work without this
    flags.set_ack_flag(True)
    req_segment = TCP_Segment(src_port, dst_port, seq_num, ack_num, flags)
    sock.sendall(req_segment.get_bytes(src_addr, dst_addr))

    # Receive Fin-Ack
    res_dgram = get_response(sock, src_port)
    res_segment = res_dgram.get_tcp_segment()

    if not fin_ack_received:
        # Send Ack packet
        flags = TCP_Flags()
        flags.set_ack_flag(True)
        seq_num = res_segment.get_ack_num()
        ack_num = res_segment.get_seq_num() + 1
        req_segment = TCP_Segment(src_port, dst_port, seq_num, ack_num, flags)
        sock.sendall(req_segment.get_bytes(src_addr, dst_addr))



def send_first_segment_last(dst_addr, dst_port, payload):
    src_port = 55555

    sock = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_TCP)
    sock.connect((dst_addr, dst_port))
    (src_addr, _) = sock.getsockname()

    cleanup()

    reset_connection(sock, src_addr, dst_addr, src_port, dst_port)

    # Needed for preventing OS from resetting TCP connection
    disable(src_port)

    (seq_num, ack_num) = establish_connection(
            sock, src_addr, dst_addr, src_port, dst_port)

    first_segment = None
    remaining_segments = []
    chunk_size = 1000
    for i in range(0, len(payload), chunk_size):
        part = payload[i:i+chunk_size]
        set_psh = len(part) < chunk_size

        # Send data
        flags = TCP_Flags()
        flags.set_ack_flag(True)
        flags.set_psh_flag(set_psh)
        req_segment = TCP_Segment(
                src_port, dst_port, seq_num, ack_num, flags, part)
        if i == 0:
            first_segment = req_segment
        else:
            remaining_segments.append(req_segment)
        seq_num += len(part)

    for req_segment in remaining_segments:
        sock.sendall(req_segment.get_bytes(src_addr, dst_addr))

    time.sleep(5) # Sleep for 5 seconds

    sock.sendall(first_segment.get_bytes(src_addr, dst_addr))

    time.sleep(5) # Wait a few seconds before closing connection

    # Receive Fin-Ack
    res_dgram = get_response(sock, src_port)
    res_segment = res_dgram.get_tcp_segment()
    seq_num = res_segment.get_ack_num()
    ack_num = res_segment.get_seq_num()

    fin_ack_received = res_segment.get_flags().get_fin_flag()

    terminate_connection(
            sock, src_addr, dst_addr, src_port, dst_port,
            seq_num, ack_num, fin_ack_received)

    sock.close()
    cleanup()

    return





import socket
import subprocess



# Need this to prevent OS from automatically responding to TCP SYN
# with a RST response.
# This is kind of a hack, and I should try to find a better
# way to mitigate the OS behavior. However, I think this is
# better than creating a routing rule.
# I first tried mitigating this by opening a dummy port;
# however, I was not able to get that to work.

def cleanup():
    # iptables -D OUTPUT -p tcp --tcp-flags RST RST -j DROP
    for i in range(0,3):
        subprocess.run([
            "iptables", "-D", "OUTPUT", "-p", "tcp",
            "--tcp-flags", "RST", "RST",
            "-j", "DROP"])

def disable(port):
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", port))
    sock.listen(0x7FFFFFFF)
    """

    subprocess.run([
        "iptables", "-A", "OUTPUT", "-p", "tcp",
        "--tcp-flags", "RST", "RST",
        "-j", "DROP"])

    def cleanup_lambda():
        # sock.close()
        cleanup()

    return cleanup_lambda





def carry_over(value):
    top_bits = value >> 16
    while top_bits > 0:
        value = (value & 0xFFFF) + top_bits
        top_bits = value >> 16
    return value

# Received help from:
# - https://gist.github.com/david-hoze/0c7021434796997a4ca42d7731a7073a
# - https://inc0x0.com/tcp-ip-packets-introduction/tcp-ip-packets-3-manually-create-and-send-raw-tcp-ip-packets/
def calculate_checksum(byte_str):
    total = 0
    for i in range(0, len(byte_str), 2):
        value = int.from_bytes(byte_str[i:i+2])
        total += value

    return 0xFFFF - carry_over(total)




from tcp_send_first_segment_last.send import send_first_segment_last


if __name__ == "__main__":
    print("Sending...")
    payload = b"A" * 0x1100
    payload += b"\n"
    # payload = b"Hello TCP.\n"

    send_first_segment_last("192.168.219.10", 1234, payload)
    # send_first_segment_last("127.0.0.1", 4444, payload)




import random
import socket
import struct

from tcp_send_first_segment_last.lib.checksum import calculate_checksum
from tcp_send_first_segment_last.lib.TCP_Segment import TCP_Segment


class IP_Datagram():
    def __init__(self, src_addr, dst_addr, tcp_segment):
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.tcp_segment = tcp_segment

    @classmethod
    def from_bytes(klass, byte_str):
        version = (byte_str[0] & 0xF0) >> 4
        IHL = (byte_str[0] & 0xF)
        total_length = int.from_bytes(byte_str[2:4])
        src_addr = int.from_bytes(byte_str[12:16])
        dst_addr = int.from_bytes(byte_str[16:20])
        tcp_segment = TCP_Segment.from_bytes(byte_str[(IHL * 4):])
        return klass(src_addr, dst_addr, tcp_segment)

    def get_tcp_segment(self):
        return self.tcp_segment

    def get_bytes(self):
        # Version and IHL
        version = 4
        IHL = 5
        ip_header_top = struct.pack(">B", (version << 4) + IHL)
        
        # DSCP and ECN
        ip_header_top += struct.pack(">B", 0)

        # Total Length
        tcp_byte_str = self.tcp_segment.get_bytes(self.src_addr, self.dst_addr)
        ip_header_top += struct.pack(">H", (IHL * 5) + len(tcp_byte_str))

        # Identification
        ip_header_top += random.randbytes(2)

        # Flags and offset
        ip_header_top += struct.pack(">H", 0)

        # Time to live
        ip_header_top += struct.pack(">B", 0x40)

        # Protocol
        ip_header_top += struct.pack(">B", 6)

        # Checksum would go here

        # Source address
        ip_header_addrs = socket.inet_aton(self.src_addr)

        # Destination address
        ip_header_addrs += socket.inet_aton(self.dst_addr)

        checksum = calculate_checksum(
                ip_header_top + struct.pack(">H", 0) + ip_header_addrs)
        ip_header = ip_header_top + struct.pack(">H", checksum) + ip_header_addrs

        return ip_header + tcp_byte_str


